#include "mccormick.h"

int    McCormick::_ITINY = 0;
double McCormick::_RTINY = 0.;

int    McCormick::_np = 0;

////////////////////////////////////////////////////////////////////////
void
McCormick::Init_McCormick
( const double l, const double u, const double cv, const double cc,
  const int icv, const int icc )
////////////////////////////////////////////////////////////////////////
{
  _sizep = _np;

  // Initialize the bounds
  _l = l;
  _u = u;

  //  Initialize the relaxations
  _cv = cv;
  _cc = cc;

  //  Initialize the relaxation subgradients
  if( _np > 0 ){
    _dcvdp = new double[_sizep];
    _dccdp = new double[_sizep];
    for ( int ip=0; ip<_np; ip++ ) _dcvdp[ip] = _dccdp[ip] = 0.;
    if ( icv >= 0 && icv < _np ) _dcvdp[icv] = 1.;
    if ( icc >= 0 && icc < _np ) _dccdp[icc] = 1.;
  }
  else{
    _dcvdp = _dccdp = NULL;
  }
}

////////////////////////////////////////////////////////////////////////
McCormick::McCormick
( const McCormick &MC )
////////////////////////////////////////////////////////////////////////
{
  _sizep = _np;

  // Initialize the bounds
  _l = MC._l;
  _u = MC._u;

  //  Initialize the relaxations
  _cv = MC._cv;
  _cc = MC._cc;

  //  Initialize the relaxation subgradients
  if( _np > 0 ){
    _dcvdp = new double[_sizep];
    _dccdp = new double[_sizep];
    for ( int ip=0; ip<_np; ip++ ){
      _dcvdp[ip] = MC._dcvdp[ip];
      _dccdp[ip] = MC._dccdp[ip];
    }
  }
  else{
    _dcvdp = _dccdp = NULL;
  }
}

////////////////////////////////////////////////////////////////////////
ostream&
operator<<
( ostream&out, const McCormick&MC)
////////////////////////////////////////////////////////////////////////
{
  out << scientific;

  // Display the bounds
  out << "\tBounds:       "
      << "[ " << MC.l() << " , " << MC.u() << " ]"  << endl;

  // Display the relaxations
  out << "\tRelaxations:  "
      << MC.cv() << "\t" << MC.cc() << endl;

  // Display the relaxation subgradients
  for( int ip=0; ip<MC._np; ip++ ){
    out << "\t              "
        << MC.dcvdp( ip ) << "\t" << MC.dccdp( ip ) << endl;
  }
  out << endl;
  
  return out;
}

////////////////////////////////////////////////////////////////////////
McCormick&
McCormick::operator=
( const double c )
////////////////////////////////////////////////////////////////////////
{
  if( _sizep != _np ) throw MC_Excp( MC_Excp::SIZE );

  // Initialize the bounds
  _l = _u = c;

  //  Initialize the relaxations
  _cv = _cc = c;
  
  //  Initialize the relaxation subgradients
  for ( int ip=0; ip<_np; ip++ ) _dcvdp[ip] = _dccdp[ip] = 0.;
  
  return *this;
}

////////////////////////////////////////////////////////////////////////
McCormick&
McCormick::operator=
( const McCormick &MCRes )
////////////////////////////////////////////////////////////////////////
{
  if( _sizep != _np ) throw MC_Excp( MC_Excp::SIZE );

  // Initialize the bounds
  _l = MCRes._l;
  _u = MCRes._u;

  //  Initialize the relaxations
  _cv = MCRes._cv;
  _cc = MCRes._cc;

  //  Initialize the relaxation subgradients
  for ( int ip=0; ip<_np; ip++ ){
    _dcvdp[ip] = MCRes._dcvdp[ip];
    _dccdp[ip] = MCRes._dccdp[ip];
  }
  
  return *this;
}

////////////////////////////////////////////////////////////////////////
McCormick
operator+
( const McCormick &MC )
////////////////////////////////////////////////////////////////////////
{
  if( MC._sizep != MC._np ) throw MC_Excp( MC_Excp::SIZE );
  McCormick MCRes( MC );

  return MCRes;
}

////////////////////////////////////////////////////////////////////////
McCormick
operator-
( const McCormick &MC )
////////////////////////////////////////////////////////////////////////
{
  if( MC._sizep != MC._np ) throw MC_Excp( MC_Excp::SIZE );
  McCormick MCRes;
  
  // Initialize the bounds
  MCRes._l = - MC._u;
  MCRes._u = - MC._l;

  //  Initialize the relaxations
  MCRes._cv = - MC._cc;
  MCRes._cc = - MC._cv;

  //  Initialize the relaxation subgradients
  for( int ip=0; ip<MCRes._np; ip++ ){
    MCRes._dcvdp[ip] = - MC._dccdp[ip];
    MCRes._dccdp[ip] = - MC._dcvdp[ip];
  }

  return MCRes;
}

////////////////////////////////////////////////////////////////////////
McCormick
operator+
( const double scalL, const McCormick &MCR )
////////////////////////////////////////////////////////////////////////
{
  if( MCR._sizep != MCR._np ) throw MC_Excp( MC_Excp::SIZE );
  McCormick MCRes;
  
  // Compute the bounds
  MCRes._l = scalL + MCR._l;
  MCRes._u = scalL + MCR._u;

  // Compute the relaxations
  MCRes._cv = scalL + MCR._cv;
  MCRes._cc = scalL + MCR._cc;

  // Compute a subgradient of the relaxations
  for( int ip=0; ip<MCRes._np; ip++ ){
    MCRes._dcvdp[ip] = MCR._dcvdp[ip];
    MCRes._dccdp[ip] = MCR._dccdp[ip];
  }

  return MCRes;
}

////////////////////////////////////////////////////////////////////////
McCormick
operator+
( const McCormick &MCL, const double scalR )
////////////////////////////////////////////////////////////////////////
{
  return scalR + MCL;
}

////////////////////////////////////////////////////////////////////////
McCormick
operator+
( const McCormick &MCL, const McCormick &MCR )
////////////////////////////////////////////////////////////////////////
{
  if( MCL._sizep != MCL._np || MCR._sizep != MCR._np ) throw MC_Excp( MC_Excp::SIZE );
  McCormick MCRes;
  
  // Compute the bounds
  MCRes._l = MCL._l + MCR._l;
  MCRes._u = MCL._u + MCR._u;

  // Compute the relaxations
  MCRes._cv = MCL._cv + MCR._cv;
  MCRes._cc = MCL._cc + MCR._cc;

  // Compute a subgradient of the relaxations
  for( int ip=0; ip<MCRes._np; ip++ ){
    MCRes._dcvdp[ip] = MCL._dcvdp[ip] + MCR._dcvdp[ip];
    MCRes._dccdp[ip] = MCL._dccdp[ip] + MCR._dccdp[ip];
  }

  return MCRes;
}

////////////////////////////////////////////////////////////////////////
McCormick
operator-
( const double scalL, const McCormick &MCR )
////////////////////////////////////////////////////////////////////////
{
  if( MCR._sizep != MCR._np ) throw MC_Excp( MC_Excp::SIZE );
  McCormick MCRes;
  
  // Compute the bounds
  MCRes._l = scalL - MCR._u;
  MCRes._u = scalL - MCR._l;

  // Compute the relaxations
  MCRes._cv = scalL - MCR._cc;
  MCRes._cc = scalL - MCR._cv;

  // Compute a subgradient of the relaxations
  for( int ip=0; ip<MCRes._np; ip++ ){
    MCRes._dcvdp[ip] = - MCR._dccdp[ip];
    MCRes._dccdp[ip] = - MCR._dcvdp[ip];
  }

  return MCRes;
}

////////////////////////////////////////////////////////////////////////
McCormick
operator-
( const McCormick &MCL, const double scalR )
////////////////////////////////////////////////////////////////////////
{
  return MCL+(-scalR);
}

////////////////////////////////////////////////////////////////////////
McCormick
operator-
( const McCormick &MCL, const McCormick &MCR )
////////////////////////////////////////////////////////////////////////
{
  if( MCL._sizep != MCL._np || MCR._sizep != MCR._np ) throw MC_Excp( MC_Excp::SIZE );
  McCormick MCRes;
  
  // Compute the bounds
  MCRes._l = MCL._l - MCR._u;
  MCRes._u = MCL._u - MCR._l;

  // Compute the relaxations
  MCRes._cv = MCL._cv - MCR._cc;
  MCRes._cc = MCL._cc - MCR._cv;

  // Compute a subgradient of the relaxations
  for( int ip=0; ip<MCRes._np; ip++ ){
    MCRes._dcvdp[ip] = MCL._dcvdp[ip] - MCR._dccdp[ip];
    MCRes._dccdp[ip] = MCL._dccdp[ip] - MCR._dcvdp[ip];
  }

  return MCRes;
}

////////////////////////////////////////////////////////////////////////
McCormick
operator*
( const double scalL, const McCormick &MCR )
////////////////////////////////////////////////////////////////////////
{
  if( MCR._sizep != MCR._np ) throw MC_Excp( MC_Excp::SIZE );
  McCormick MCRes;

  // CASE: multiplying by a positive scalar value
  if ( scalL >= 0 ){
  
    // Compute the bounds
    MCRes._l = scalL * MCR._l;
    MCRes._u = scalL * MCR._u;

    // Compute the relaxations
    MCRes._cv = scalL * MCR._cv;
    MCRes._cc = scalL * MCR._cc;

    // Compute a subgradient of the relaxations
    for( int ip=0; ip<MCRes._np; ip++ ){
      MCRes._dcvdp[ip] = scalL * MCR._dcvdp[ip];
      MCRes._dccdp[ip] = scalL * MCR._dccdp[ip];
    }
  }

  // CASE: multiplying by a negative scalar value
  else{
  
    // Compute the bounds
    MCRes._l = scalL * MCR._u;
    MCRes._u = scalL * MCR._l;

    // Compute the relaxations
    MCRes._cv = scalL * MCR._cc;
    MCRes._cc = scalL * MCR._cv;

    // Compute a subgradient of the relaxations
    for( int ip=0; ip<MCRes._np; ip++ ){
      MCRes._dcvdp[ip] = scalL * MCR._dccdp[ip];
      MCRes._dccdp[ip] = scalL * MCR._dcvdp[ip];
    }
  }

  return MCRes;
}

////////////////////////////////////////////////////////////////////////
McCormick
operator*
( const McCormick &MCL, const double scalR )
////////////////////////////////////////////////////////////////////////
{  
  return scalR*MCL;
}

////////////////////////////////////////////////////////////////////////
McCormick
operator*
( const McCormick &MCL, const McCormick &MCR )
////////////////////////////////////////////////////////////////////////
{
  // CASE: a_u >= 0
  if ( MCL._l >= 0. ){

    // SUBCASE: a_v >= 0
    if ( MCR._l >= 0. ){

      if( MCL._sizep != MCL._np || MCR._sizep != MCR._np )
        throw MC_Excp( MC_Excp::SIZE );
      McCormick MCRes;

      // Compute the bounds
      MCRes._l = MCL._l * MCR._l;
      MCRes._u = MCL._u * MCR._u;

      // Compute the convex underestimator and a subgradient of it
      double cv1 = MCR._u * MCL._cv + MCL._u * MCR._cv - MCL._u * MCR._u;
      double cv2 = MCR._l * MCL._cv + MCL._l * MCR._cv - MCL._l * MCR._l;
      //if ( cv1 - cv2 > MCRes._RTINY * fabs(cv1 + cv2) + MCRes._RTINY ){
      if ( cv1 > cv2 ){
        MCRes._cv = cv1;
        for( int ip=0; ip<MCRes._np; ip++ )
          MCRes._dcvdp[ip] = MCR._u * MCL._dcvdp[ip] + MCL._u * MCR._dcvdp[ip];
      }
      else{
        MCRes._cv = cv2;
        for( int ip=0; ip<MCRes._np; ip++ )
          MCRes._dcvdp[ip] = MCR._l * MCL._dcvdp[ip] + MCL._l * MCR._dcvdp[ip];
      }

      // Compute the concave overestimator and a subgradient of it
      double cc1 = MCR._l * MCL._cc + MCL._u * MCR._cc - MCL._u * MCR._l;
      double cc2 = MCR._u * MCL._cc + MCL._l * MCR._cc - MCL._l * MCR._u;
      //if ( cc1 - cc2 < MCRes._RTINY * fabs(cc1 + cc2) + MCRes._RTINY ){
      if ( cc1 < cc2 ){
        MCRes._cc = cc1;
        for( int ip=0; ip<MCRes._np; ip++ )
          MCRes._dccdp[ip] = MCR._l * MCL._dccdp[ip] + MCL._u * MCR._dccdp[ip];
      }
      else{
        MCRes._cc = cc2;
        for( int ip=0; ip<MCRes._np; ip++ )
          MCRes._dccdp[ip] = MCR._u * MCL._dccdp[ip] + MCL._l * MCR._dccdp[ip];
      }

      return MCRes;
    }

    // SUBCASE: b_v <= 0
    else if ( MCR._u <= 0. ) return -( MCL * (-MCR) );

    // SUBCASE: a_v < 0 < b_v
    else{

      if( MCL._sizep != MCL._np || MCR._sizep != MCR._np )
        throw MC_Excp( MC_Excp::SIZE );
      McCormick MCRes;

      // Compute the bounds
      MCRes._l = MCL._u * MCR._l;
      MCRes._u = MCL._u * MCR._u;

      // Compute the convex underestimator and a subgradient of it
      double cv1 = MCR._u * MCL._cv + MCL._u * MCR._cv - MCL._u * MCR._u;
      double cv2 = MCR._l * MCL._cc + MCL._l * MCR._cv - MCL._l * MCR._l;
      if ( cv1 > cv2 ){
      //if ( cv1 - cv2 > MCRes._RTINY * fabs(cv1 + cv2) + MCRes._RTINY ){
        MCRes._cv = cv1;
        for( int ip=0; ip<MCRes._np; ip++ )
          MCRes._dcvdp[ip] = MCR._u * MCL._dcvdp[ip] + MCL._u * MCR._dcvdp[ip];
      }
      else{
        MCRes._cv = cv2;
        for( int ip=0; ip<MCRes._np; ip++ )
          MCRes._dcvdp[ip] = MCR._l * MCL._dccdp[ip] + MCL._l * MCR._dcvdp[ip];
      }

      // Compute the concave overestimator and a subgradient of it
      double cc1 = MCR._l * MCL._cv + MCL._u * MCR._cc - MCL._u * MCR._l;
      double cc2 = MCR._u * MCL._cc + MCL._l * MCR._cc - MCL._l * MCR._u;
      if ( cc1 < cc2 ){
      //if ( cc1 - cc2 < MCRes._RTINY * fabs(cc1 + cc2) + MCRes._RTINY ){
        MCRes._cc = cc1;
        for( int ip=0; ip<MCRes._np; ip++ )
          MCRes._dccdp[ip] = MCR._l * MCL._dcvdp[ip] + MCL._u * MCR._dccdp[ip];
      }
      else{
        MCRes._cc = cc2;
        for( int ip=0; ip<MCRes._np; ip++ )
          MCRes._dccdp[ip] = MCR._u * MCL._dccdp[ip] + MCL._l * MCR._dccdp[ip];
      }

      return MCRes;
    }
  }
  
  // CASE: b_u <= 0
  else if ( MCL._u <= 0. ){

    // SUBCASE: a_v >= 0
    if ( MCR._l >= 0. ) return -( (-MCL) * MCR);

    // SUBCASE: b_v <= 0
    else if ( MCR._u <= 0. ) return (-MCL) * (-MCR);

    // SUBCASE: a_v < 0 < b_v
    else return -( MCR * (-MCL) );
  }

  // CASE: a_u < 0 < b_u
  else{

    // SUBCASE: a_v >= 0
    if ( MCR._l >= 0. ) return MCR * MCL;

    // SUBCASE: b_v <= 0
    else if ( MCR._u <= 0. ) return -( (-MCR) * MCL );

    // SUBCASE: a_v < 0 < b_v
    else{

      if( MCL._sizep != MCL._np || MCR._sizep != MCR._np )
        throw MC_Excp( MC_Excp::SIZE );
      McCormick MCRes;

      // Compute the bounds
      MCRes._l = min( MCL._u * MCR._l, MCL._l * MCR._u );
      MCRes._u = max( MCL._l * MCR._l, MCL._u * MCR._u );

      // Compute the convex underestimator and a subgradient of it
      double cv1 = MCR._u * MCL._cv + MCL._u * MCR._cv - MCL._u * MCR._u;
      double cv2 = MCR._l * MCL._cc + MCL._l * MCR._cc - MCL._l * MCR._l;
      if ( cv1 > cv2 ){
      //if ( cv1 - cv2 > MCRes._RTINY * fabs(cv1 + cv2) + MCRes._RTINY ){
        MCRes._cv = cv1;
        for( int ip=0; ip<MCRes._np; ip++ )
          MCRes._dcvdp[ip] = MCR._u * MCL._dcvdp[ip] + MCL._u * MCR._dcvdp[ip];
      }
      else{
        MCRes._cv = cv2;
        for( int ip=0; ip<MCRes._np; ip++ )
          MCRes._dcvdp[ip] = MCR._l * MCL._dccdp[ip] + MCL._l * MCR._dccdp[ip];
      }

      // Compute the concave overestimator and a subgradient of it
      double cc1 = MCR._l * MCL._cv + MCL._u * MCR._cc - MCL._u * MCR._l;
      double cc2 = MCR._u * MCL._cc + MCL._l * MCR._cv - MCL._l * MCR._u;
      if ( cc1 < cc2 ){
      //if ( cc1 - cc2 < MCRes._RTINY * fabs(cc1 + cc2) + MCRes._RTINY ){
        MCRes._cc = cc1;
        for( int ip=0; ip<MCRes._np; ip++ )
          MCRes._dccdp[ip] = MCR._l * MCL._dcvdp[ip] + MCL._u * MCR._dccdp[ip];
      }
      else{
        MCRes._cc = cc2;
        for( int ip=0; ip<MCRes._np; ip++ )
          MCRes._dccdp[ip] = MCR._u * MCL._dccdp[ip] + MCL._l * MCR._dcvdp[ip];
      }

      return MCRes;
    }
  }

  // error if the code terminates here!!!
  // return object with 0 values and relaxations
  McCormick MCRes;
  return MCRes;
}

////////////////////////////////////////////////////////////////////////
McCormick
operator/
( const McCormick &MCL, const double scalR )
////////////////////////////////////////////////////////////////////////
{
  // Check that division is safe
  if ( McCormick::_isequal( scalR, 0. )) throw MC_Excp( MC_Excp::DIV );

//   if ( scalR == 0. ){
//     cerr << "Fatal error: Division by 0!" << endl << "program terminates..." << endl;
//     exit(-1);
//   }

  return  ( 1. / scalR ) * MCL;
}

////////////////////////////////////////////////////////////////////////
McCormick
operator/
( const double scalL, const McCormick &MCR )
////////////////////////////////////////////////////////////////////////
{
  return scalL * inv( MCR );
}

////////////////////////////////////////////////////////////////////////
McCormick
operator/
( const McCormick &MCL, const McCormick &MCR )
////////////////////////////////////////////////////////////////////////
{
  return MCL * inv( MCR );
}

////////////////////////////////////////////////////////////////////////
McCormick
inv
( const McCormick &MC )
////////////////////////////////////////////////////////////////////////
{
  // check that inversion is safe
  if ( MC._l <= 0. && MC._u >= 0. ) throw MC_Excp( MC_Excp::INV );

  if( MC._sizep != MC._np ) throw MC_Excp( MC_Excp::SIZE );
  McCormick MCRes;

  // Compute the bounds
  MCRes._l = 1. / MC._u;
  MCRes._u = 1. / MC._l;

  // CASE: a_u > 0
  if ( MC._l > 0. ){

    // Compute the convex underestimator and a subgradient of it
    { int indexMid = -1;
      double valMid = mid( MC._cv, MC._cc, MC._u, indexMid );
      MCRes._cv = 1. / valMid;
      for( int ip=0; ip<MCRes._np; ip++ )
        MCRes._dcvdp[ip] = - mid( MC._dcvdp, MC._dccdp, ip, indexMid )
          / ( valMid * valMid );
    }

    // Compute the concave overestimator and a subgradient of it
    { int indexMid = -1;
      MCRes._cc = 1. / MC._l + 1. / MC._u - mid( MC._cv, MC._cc, MC._l, indexMid )
        / ( MC._l * MC._u );
      for( int ip=0; ip<MCRes._np; ip++ )
        MCRes._dccdp[ip] = - mid( MC._dcvdp, MC._dccdp, ip, indexMid )
          / ( MC._l * MC._u );
    }
  }

  // CASE: b_u < 0
  else{

    // Compute the convex underestimator and a subgradient of it
    { int indexMid = -1;
      MCRes._cv = 1. / MC._l + 1. / MC._u - mid( MC._cv, MC._cc, MC._u, indexMid )
        / ( MC._l * MC._u );
      for( int ip=0; ip<MCRes._np; ip++ )
        MCRes._dcvdp[ip] = - mid( MC._dcvdp, MC._dccdp, ip, indexMid )
          / ( MC._l * MC._u );
    }

    // Compute the concave overestimator and a subgradient of it
    { int indexMid = -1;
      double valMid = mid( MC._cv, MC._cc, MC._l, indexMid);
      MCRes._cc = 1. / valMid;
      for( int ip=0; ip<MCRes._np; ip++ )
        MCRes._dccdp[ip] = - mid( MC._dcvdp, MC._dccdp, ip, indexMid )
          / ( valMid * valMid );
    }
  }
  
  return MCRes;
}

////////////////////////////////////////////////////////////////////////
McCormick
exp
( const McCormick &MC )
////////////////////////////////////////////////////////////////////////
{
  if( MC._sizep != MC._np ) throw MC_Excp( MC_Excp::SIZE );
  McCormick MCRes;

  // Compute the bounds
  MCRes._l = exp( MC._l );
  MCRes._u = exp( MC._u );
  
  // Compute the convex underestimator and a subgradient of it
  { int indexMid = -1;
    MCRes._cv = exp( mid( MC._cv, MC._cc, MC._l, indexMid ));
    for( int ip=0; ip<MCRes._np; ip++ )
      MCRes._dcvdp[ip] = mid( MC._dcvdp, MC._dccdp, ip, indexMid ) * MCRes._cv;
  }

  // Compute the concave overestimator and a subgradient of it
  { int indexMid = -1;
    double scal = 0.;
    if( !McCormick::_isequal( MC._l, MC._u ))
      scal = ( MCRes._u - MCRes._l ) / ( MC._u - MC._l );
    MCRes._cc = exp( MC._l ) + scal * ( mid( MC._cv, MC._cc, MC._u, indexMid )
      - MC._l );
    for( int ip=0; ip<MCRes._np; ip++ )
      MCRes._dccdp[ip] = mid( MC._dcvdp, MC._dccdp, ip, indexMid ) * scal;
  }

  return MCRes;
}

////////////////////////////////////////////////////////////////////////
McCormick
arh
( const McCormick &MC, const double constArreh )
////////////////////////////////////////////////////////////////////////
{
  // check that applying the Arrehenius-like function is safe
  if ( MC._l <= 0. || constArreh < 0. ) throw MC_Excp( MC_Excp::ARH );

  // CASE: TU <= k/2 : convex portion
  if ( MC._u <= constArreh / 2. ){

    if( MC._sizep != MC._np ) throw MC_Excp( MC_Excp::SIZE );
    McCormick MCRes;

    // Compute the bounds
    MCRes._l = exp( - constArreh / MC._l );
    MCRes._u = exp( - constArreh / MC._u );
  
    // Compute the convex underestimator and a subgradient of it
    { int indexMid = -1;
      double valMid = mid( MC._cv, MC._cc, MC._l, indexMid );
      MCRes._cv = exp( - constArreh / valMid );
      for( int ip=0; ip<MCRes._np; ip++ )
        MCRes._dcvdp[ip] = constArreh / ( valMid * valMid ) * MCRes._cv
          * mid( MC._dcvdp, MC._dccdp, ip, indexMid );
    }

    // Compute the concave overestimator and a subgradient of it
    { int indexMid = -1;
      double scal = 0.;
      if( !McCormick::_isequal( MC._l, MC._u ))
        scal = ( MCRes._u - MCRes._l ) / ( MC._u - MC._l );
      MCRes._cc = MCRes._l + scal * ( mid( MC._cv, MC._cc, MC._u, indexMid )
        - MC._l );
      for( int ip=0; ip<MCRes._np; ip++ )
        MCRes._dccdp[ip] = mid( MC._dcvdp, MC._dccdp, ip, indexMid ) * scal;
    }
    return MCRes;
  }
  
  // CASE: TL >= k/2 : concave portion
  else if ( MC._l >= constArreh / 2. ){

    if( MC._sizep != MC._np ) throw MC_Excp( MC_Excp::SIZE );
    McCormick MCRes;

    // Compute the bounds
    MCRes._l = exp( - constArreh / MC._l );
    MCRes._u = exp( - constArreh / MC._u );
  
    // Compute the convex underestimator and a subgradient of it
    { int indexMid = -1;
      double scal = 0.;
      if( !McCormick::_isequal( MC._l, MC._u ))
        scal = ( MCRes._u - MCRes._l ) / ( MC._u - MC._l );
      MCRes._cv = MCRes._l + scal * ( mid( MC._cv, MC._cc, MC._l, indexMid )
        - MC._l );
      for( int ip=0; ip<MCRes._np; ip++ )
        MCRes._dcvdp[ip] = mid( MC._dcvdp, MC._dccdp, ip, indexMid ) * scal;
    }

    // Compute the concave overestimator and a subgradient of it
    { int indexMid = -1;
      double valMid = mid( MC._cv, MC._cc, MC._u, indexMid );
      MCRes._cc = exp( - constArreh / valMid );
      for( int ip=0; ip<MCRes._np; ip++ )
        MCRes._dccdp[ip] = constArreh / ( valMid * valMid ) * MCRes._cc
          * mid( MC._dcvdp, MC._dccdp, ip, indexMid );
    }
    return MCRes;
  }
  
  // CASE ELSE: mixed convex/concave portion
  else
    return exp( - constArreh * inv( MC ) );

}

////////////////////////////////////////////////////////////////////////
McCormick
log
( const McCormick &MC )
////////////////////////////////////////////////////////////////////////
{
  // check that applying the log function is safe
  if ( MC._l <= 0. ) throw MC_Excp( MC_Excp::LOG );

  if( MC._sizep != MC._np ) throw MC_Excp( MC_Excp::SIZE );
  McCormick MCRes;

  // Compute the bounds
  MCRes._l = log( MC._l );
  MCRes._u = log( MC._u );

  // Compute the convex underestimator and a subgradient of it
  { int indexMid = -1;
    double scal = 0.;
    if( !McCormick::_isequal( MC._l, MC._u ))
      scal = ( MCRes._u - MCRes._l ) / ( MC._u - MC._l );
    MCRes._cv = MCRes._l + scal * ( mid( MC._cv, MC._cc, MC._l, indexMid )
      - MC._l );
    for( int ip=0; ip<MCRes._np; ip++ )
      MCRes._dcvdp[ip] = mid( MC._dcvdp, MC._dccdp, ip, indexMid ) * scal;
  }

  // Compute the concave overestimator and a subgradient of it
  { int indexMid = -1;
    double valMid = mid( MC._cv, MC._cc, MC._u, indexMid );
    MCRes._cc = log( valMid );
    for( int ip=0; ip<MCRes._np; ip++ )
      MCRes._dccdp[ip] = mid( MC._dcvdp, MC._dccdp, ip, indexMid ) / valMid;
  }

  return MCRes;
}

////////////////////////////////////////////////////////////////////////
McCormick
xlog
( const McCormick &MC )
////////////////////////////////////////////////////////////////////////
{
  // check that applying the log function is safe
  if ( MC._l <= 0. ) throw MC_Excp( MC_Excp::LOG );

  if( MC._sizep != MC._np ) throw MC_Excp( MC_Excp::SIZE );
  McCormick MCRes;

  // Compute the bounds
  double valL, valU;
  { int indexMid = -1;
    valL = mid( MC._l, MC._u, exp(-1.), indexMid );
    MCRes._l = xlog(valL);
    valU = ( xlog(MC._u)>=xlog(MC._l)? MC._u: MC._l );
    MCRes._u = xlog(valU);
  }

  // Compute the convex underestimator and a subgradient of it
  { int indexMid = -1;
    double valMid = mid( MC._cv, MC._cc, valL, indexMid );
    MCRes._cv = xlog( valMid );
    for( int ip=0; ip<MCRes._np; ip++ )
      MCRes._dcvdp[ip] = (log( valMid ) + 1.) * mid( MC._dcvdp, MC._dccdp, ip, indexMid );
  }

  // Compute the concave overestimator and a subgradient of it
  { int indexMid = -1;
    double scal = 0.;
    if( !McCormick::_isequal( MC._l, MC._u ))
      scal = ( xlog(MC._u) - xlog(MC._l) ) / ( MC._u - MC._l );
    MCRes._cc = xlog(MC._l) + scal * ( mid( MC._cv, MC._cc, valU, indexMid ) - MC._l );
    for( int ip=0; ip<MCRes._np; ip++ )
      MCRes._dccdp[ip] = mid( MC._dcvdp, MC._dccdp, ip, indexMid ) * scal;
  }

  return MCRes;
}

////////////////////////////////////////////////////////////////////////
McCormick
sqrt
( const McCormick &MC )
////////////////////////////////////////////////////////////////////////
{
  // check that applying the sqrt function is safe
  if ( MC._l < 0. ) throw MC_Excp( MC_Excp::SQRT );

  if( MC._sizep != MC._np ) throw MC_Excp( MC_Excp::SIZE );
  McCormick MCRes;

  // Compute the bounds
  MCRes._l = sqrt( MC._l );
  MCRes._u = sqrt( MC._u );

  // Compute the convex underestimator and a subgradient of it
  { double scal = 0.;
    if( !McCormick::_isequal( MC._l, MC._u ))
      scal = ( MCRes._u - MCRes._l ) / ( MC._u - MC._l );
    int indexMid = -1;
    double valMid = mid( MC._cv, MC._cc, MC._l, indexMid );
    MCRes._cv = MCRes._l + scal * ( valMid - MC._l );
    for( int ip=0; ip<MCRes._np; ip++ )
      MCRes._dcvdp[ip] = mid( MC._dcvdp, MC._dccdp, ip, indexMid ) * scal;
  }

  // Compute the concave overestimator and a subgradient of it
  { int indexMid = -1;
    double valMid = mid( MC._cv, MC._cc, MC._u, indexMid );
    MCRes._cc = sqrt( valMid );
    for( int ip=0; ip<MCRes._np; ip++ )
      MCRes._dccdp[ip] = mid( MC._dcvdp, MC._dccdp, ip, indexMid ) / (2.*valMid);
  }

  return MCRes;
}

////////////////////////////////////////////////////////////////////////
McCormick
fabs
( const McCormick &MC )
////////////////////////////////////////////////////////////////////////
{
  if( MC._sizep != MC._np ) throw MC_Excp( MC_Excp::SIZE );
  McCormick MCRes;

  // Compute the bounds
  { int indexMid = -1;
    MCRes._l = fabs( mid( MC._l, MC._u, 0., indexMid ));
    MCRes._u = max( fabs(MC._l), fabs(MC._u) );
  }

  // Compute the convex underestimator and a subgradient of it
  { int indexMid = -1;
    double zmin = mid( MC._l, MC._u, 0., indexMid );
    indexMid = -1;
    double valmid = mid( MC._cv, MC._cc, zmin, indexMid );
    MCRes._cv = fabs( valmid );
    if( valmid >= 0. ){
      for( int ip=0; ip<MCRes._np; ip++ ){
        MCRes._dcvdp[ip] = mid( MC._dcvdp, MC._dccdp, ip, indexMid );
      }
    }
    else{
      for( int ip=0; ip<MCRes._np; ip++ ){
        MCRes._dcvdp[ip] = - mid( MC._dcvdp, MC._dccdp, ip, indexMid );
      }
    }
  }

  // Compute the concave overestimator and a subgradient of it
  { int indexMid = -1;
    double zmax = (fabs( MC._l )>fabs( MC._u )? MC._l: MC._u);
    double scal = ( McCormick::_isequal( MC._l, MC._u )? 0.:
      ( fabs( MC._u ) - fabs( MC._l ) ) / ( MC._u - MC._l ) );
    MCRes._cc = fabs( MC._l ) + scal * ( mid( MC._cv, MC._cc,
      zmax, indexMid ) - MC._l );
    for( int ip=0; ip<MCRes._np; ip++ )
      MCRes._dccdp[ip] = mid( MC._dcvdp, MC._dccdp, ip, indexMid ) * scal;
  }

  return MCRes;
}

////////////////////////////////////////////////////////////////////////
McCormick
pow
( const McCormick &MC, const int intExponent )
////////////////////////////////////////////////////////////////////////
{
  // CASE: exponent = 0
  if ( intExponent == 0 ){
    McCormick MCRes( 1. );
    return MCRes;
  }

  // CASE: exponent = 1
  else if ( intExponent == 1 ){
    McCormick MCRes( MC );
    return MCRes;
  }

  // CASE: exponent >= 2
  else if ( intExponent >= 2 ){

    // SUBCASE: exponent is even number
    if ( intExponent%2 == 0 ){

      if( MC._sizep != MC._np ) throw MC_Excp( MC_Excp::SIZE );
      McCormick MCRes;

      // Compute the bounds
      { int indexMid = -1;
        MCRes._l = pow( mid( MC._l, MC._u, 0., indexMid ), intExponent );
        MCRes._u = max( pow(MC._l,intExponent), pow(MC._u,intExponent) );
      }

      // Compute the convex underestimator and a subgradient of it
      { int indexMid = -1;
        double zmin = mid( MC._l, MC._u, 0., indexMid );
        indexMid = -1;
        MCRes._cv = pow( mid( MC._cv, MC._cc, zmin, indexMid ), intExponent );
        for( int ip=0; ip<MCRes._np; ip++ ){
          MCRes._dcvdp[ip] = intExponent * mid( MC._dcvdp, MC._dccdp, ip, indexMid )
            * pow( mid( MC._cv, MC._cc, MC._l, indexMid ), intExponent-1 );
        }
      }

      // Compute the concave overestimator and a subgradient of it
      { int indexMid = -1;
        double zmax = (pow( MC._l, intExponent )>pow( MC._u, intExponent )? MC._l: MC._u);
        double scal = ( McCormick::_isequal( MC._l, MC._u )? 0.:
          ( pow( MC._u, intExponent ) - pow( MC._l, intExponent ) ) / ( MC._u - MC._l ) );
        MCRes._cc = pow( MC._l, intExponent ) + scal * ( mid( MC._cv, MC._cc,
          zmax, indexMid ) - MC._l );
        for( int ip=0; ip<MCRes._np; ip++ )
          MCRes._dccdp[ip] = mid( MC._dcvdp, MC._dccdp, ip, indexMid ) * scal;
      }

      return MCRes;
    }

    // SUBCASE: exponent is odd number
    else return ( pow( MC, intExponent-1 ) * MC );

  } // ends CASE: exponent >= 2

  // CASE: exponent <= -1
  else return ( inv ( pow( MC, -intExponent )));
}

////////////////////////////////////////////////////////////////////////
McCormick
pow
( const McCormick &MC, const double realExponent )
////////////////////////////////////////////////////////////////////////
{
  return exp( realExponent * log( MC ) );
}

////////////////////////////////////////////////////////////////////////
McCormick
pow
( const McCormick &MC, const McCormick &MCExp )
////////////////////////////////////////////////////////////////////////
{
  return exp( MCExp * log( MC ) );
}

////////////////////////////////////////////////////////////////////////
bool
McCormick::_isequal
( const double real1, const double real2 )
////////////////////////////////////////////////////////////////////////
{
  return _isequal( real1, real2, _machprec(), _machprec() );
}

////////////////////////////////////////////////////////////////////////
bool
McCormick::_isequal
( const double real1, const double real2, const double atol,
  const double rtol )
////////////////////////////////////////////////////////////////////////
{
  // compare values
  double gap = fabs(real1-real2);
  double ave = 0.5*fabs(real1+real2);
  return (gap > atol + ave*rtol)? false: true;
}

////////////////////////////////////////////////////////////////////////
double
McCormick::_machprec()
////////////////////////////////////////////////////////////////////////
{
  if( _ITINY ) return _RTINY;

  double dummy = 1.0, done = 1.0;
  while ( done + dummy > done ) dummy /= 2.0;
  _ITINY = 1;
  _RTINY = 2.*dummy;

  return _RTINY;
}
