#ifndef MCCORMICK_H
#define MCCORMICK_H
#include <iostream>
#include <vector>
#include <cmath>

using namespace std;

extern double mid( const double, const double, const double, int& );
extern double mid( const double*, const double*, const int, const int );
extern double arh( const double, const double );
extern double xlog( const double );

//! @brief C++ class for handling exceptions during the McCormick relaxation of a nonconvex function on a box.
////////////////////////////////////////////////////////////////////////
class MC_Excp
////////////////////////////////////////////////////////////////////////
{
public:

  //! @brief Enumeration type for McCormick exception handling
  enum MC_Err{
    SIZE=1,	//!< The size of the problem has changed (number of parameters)
    DIV,	//!< Error during the relaxation of a division term (division by zero)
    INV,	//!< Error during the relaxation of an inverse function (zero in range)
    LOG,	//!< Error during the relaxation of a logarithm function (negative values in range)
    SQRT,	//!< Error during the relaxation of a squre-root function (negative values in range)
    ARH 	//!< Error during the relaxation of an Arrhenius-like function (negative values in range, or negative coefficient)
  };

  //! @brief Constructor for error <a>ierr</a>
  MC_Excp( MC_Err ierr ) : _ierr( ierr ){}

  //! @brief Inline function returning the error flag
  int ierr(){ return _ierr; }

private:

  MC_Err _ierr;
};


//! @brief C++ class doing the McCormick relaxation of a nonconvex function on a box.
////////////////////////////////////////////////////////////////////////
//! McCormick is an C++ class for calculating the McCormick relaxation
//! of a nonconvex function on a box, as well as valid subgradients of
//! these relaxations. Any function can be considered, provided it is
//! factorable.
////////////////////////////////////////////////////////////////////////
class McCormick
////////////////////////////////////////////////////////////////////////
{
  // friends of class McCormick for operator overloading
  friend McCormick operator+ ( const McCormick& );
  friend McCormick operator+ ( const McCormick&, const McCormick& );
  friend McCormick operator+ ( const double, const McCormick& );
  friend McCormick operator+ ( const McCormick&, const double );
  friend McCormick operator- ( const McCormick& );
  friend McCormick operator- ( const McCormick&, const McCormick& );
  friend McCormick operator- ( const double, const McCormick& );
  friend McCormick operator- ( const McCormick&, const double );
  friend McCormick operator* ( const McCormick&, const McCormick& );
  friend McCormick operator* ( const McCormick&, const double );
  friend McCormick operator* ( const double, const McCormick& );
  friend McCormick operator/ ( const McCormick&, const McCormick& );
  friend McCormick operator/ ( const McCormick&, const double );
  friend McCormick operator/ ( const double, const McCormick& );
  friend ostream& operator<< ( ostream&, const McCormick& );

  // friends of class McCormick for function overloading
  friend McCormick inv  ( const McCormick& );
  friend McCormick exp  ( const McCormick& );
  friend McCormick log  ( const McCormick& );
  friend McCormick fabs ( const McCormick& );
  friend McCormick sqrt ( const McCormick& );
  friend McCormick xlog ( const McCormick& );
  friend McCormick arh  ( const McCormick&, const double );
  friend McCormick pow  ( const McCormick&, const int );
  friend McCormick pow  ( const McCormick&, const double );
  friend McCormick pow  ( const McCormick&, const McCormick& );
  friend double mid( const double, const double, const double, int& );
  friend double mid( const double*, const double*, const int, const int );
  
public:

  // other operator overloadings
  McCormick& operator=
    ( const McCormick& );
  McCormick& operator=
    ( const double );

  /** @defgroup MCCORMICK McCormick Relaxations
   *  @{
   */
  //! @brief Constructor for a constant value <a>c</a>.
  McCormick
    ( const double c )
    { Init_McCormick( c, c, c, c ); };
  //! @brief Constructor for a variable that belongs to the interval [<a>l</a>,<a>u</a>], has the value <a>c</a>, and has index <a>ip</a>.
  McCormick
    ( const double l, const double u, const double c, const int ip=-1 )
    { Init_McCormick( l, u, c, c, ip, ip ); };
  //! @brief Constructor for an intermediate variable, that belongs to the interval [<a>l</a>,<a>u</a>], has convex underestimator value <a>cv</a> with index <a>icv</a>, and concave overestimator <a>cc</a> with index <a>icc</a>.
  McCormick
    ( const double l, const double u, const double cv, const double cc,
      const int icv=-1, const int icc=-1 )
    { Init_McCormick( l, u, cv, cc, icv, icc ); };
  //! @brief Copy constructor
  McCormick
    ( const McCormick& );

  //! @brief Destructor
  ~McCormick();
  // private constructor
  McCormick();

  //! @brief Returns the number of variables in the original function.
  int np() const
    { return _np; };
  //! @brief Returns the lower bounding value.
  double l() const
    { return _l;  };
  //! @brief Returns the upper bounding value.
  double u() const
    { return _u;  };
  //! @brief Returns the value of the McCormick's convex underestimator.
  double cv() const
    { return _cv; };
  //! @brief Returns the value of the McCormick's concave overestimator.
  double cc() const
    { return _cc; };
  //! @brief Returns a pointer to a subgradient of the McCormick's convex underestimator.
  const double* dcvdp() const
    { return _dcvdp; };
  //! @brief Returns a pointer to a subgradient of the McCormick's concave overestimator.
  const double* dccdp() const
    { return _dccdp; };
  //! @brief Returns the ip-th component of a subgradient of the McCormick's convex underestimator.
  double dcvdp( const int ip ) const
    { return _dcvdp[ip]; };
  //! @brief Returns the ip-th component of a subgradient of the McCormick's concave overestimator.
  double dccdp( const int ip ) const
    { return _dccdp[ip]; };

  //! @brief Sets the lower bounding value.
  void l ( double l )  { _l = l; };
  //! @brief Sets the upper bounding value.
  void u ( double u )  { _u = u; };
  //! @brief Sets the value of the McCormick's convex underestimator.
  void cv( double cv ) { _cv = cv; };
  //! @brief Sets the value of the McCormick's concave overestimator.
  void cc( double cc ) { _cc = cc; };

  //! @brief Sets the number of variables in the original function; needed to calculate a subgradient of a McCormick relaxation.
  static void np( int npar ) { _np = npar; };
  /** @} */
  
private:

  static int _ITINY;
  static double _RTINY;

  //! @brief Number of variables (same for all objects)
  static int _np;
  //! @brief Number of variables in actual object
  int _sizep;
  //! @brief Lower bound
  double _l;
  //! @brief Upper bound
  double _u;
  //! @brief McCormick's convex underestimator
  double _cv;
  //! @brief McCormick's concave overestimator
  double _cc;
  //! @brief Subgradient of McCormick's convex underestimator
  double *_dcvdp;
  //! @brief Subgradient of McCormick's concave overestimator
  double *_dccdp;

  //! @brief Initializes a McCormick class object
  void Init_McCormick
   ( const double, const double, const double, const double,
     const int=-1, const int=-1 );

  //! @brief Calculates the machine precision for 'double' data type
  static double _machprec();
  //! @brief Tests whether two real values are identical modulo some specified absolute and relative tolerances
  static bool _isequal
    ( const double real1, const double real2, const double atol,
      const double rtol );
  //! @brief Tests whether two real values are identical modulo the machine precision
  static bool _isequal
    ( const double real1, const double real2 );
};

////////////////////////////////////////////////////////////////////////
inline McCormick::McCormick()
: _l( 0. ), _u( 0. ), _cv( 0. ), _cc( 0. )
////////////////////////////////////////////////////////////////////////
{
  _sizep = _np;

  // simply allocate the arrays without initializing the values
  if( _sizep > 0 ){
    _dcvdp = new double[ _sizep ];
    _dccdp = new double[ _sizep ];
  }
  else{
    _dcvdp = _dccdp = NULL;
  }
}

////////////////////////////////////////////////////////////////////////
inline McCormick::~McCormick()
////////////////////////////////////////////////////////////////////////
{
  delete [] _dcvdp;
  delete [] _dccdp;
}

#endif
